<?php

namespace Drupal\odt_importer;

use \XMLReader;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Defines an ODT to HTML conversion implementation.
 *
 */

/**
 * Global ToDos: 
 * - Parameter and return typehinting for class methods:
 * https://www.drupal.org/docs/develop/standards/coding-standards#s-parameter-and-return-typehinting
 * - Create templates for elements for easy overriding markup in themes
 */

class OdtConverter {
  use MessengerTrait;
  use StringTranslationTrait;

  /**
   * Features enabled for odt converter.
   *
   * @var array
   */
  protected $features;

  /**
   * Folder for storing images from odt file.
   *
   * @var string
   */
  protected $images_folder;

  /**
   * Converted html.
   *
   * @var array
   */
  protected $html;

  /**
   * Rendered footnotes.
   *
   * @var array
   */
  protected $footnotes;

  /**
   * {@inheritdoc}
   */
  public function __construct($config, $odt_file) {
    // Configuration settings.
    //  FALSE : do not parse, do not print
    //  TRUE : print  and apply all supported HTML tags and styles
    //  (REMOVED !!!Recheck this!!!) 1 : print as simple text (do not apply any HTML tag or style)
    //  ToDo: Add this settings to Drupal configurtion GUI
    $this->features['header'] = (bool) $config['features']['header'] ?? TRUE;
    $this->features['list'] = (bool) $config['features']['list'] ?? TRUE;
    $this->features['table'] = (bool) $config['features']['table'] ?? TRUE;
    $this->features['link'] = (bool) $config['features']['link'] ?? TRUE;
    $this->features['frame'] = (bool) $config['features']['frame'] ?? TRUE;
    $this->features['image'] = (bool) $config['features']['image'] ?? TRUE;
    $this->features['note'] = (bool) $config['features']['note'] ?? TRUE;
    $this->features['annotation'] = (bool) $config['features']['annotation'] ?? TRUE;
    $this->features['toc'] = (bool) $config['features']['toc'] ?? TRUE;
    $this->features['bookmark'] = (bool) $config['features']['bookmark'] ?? TRUE;

    // Do we need this option? Since we generate footnotes with annotations enabled
    //$this->features['footnote'] = (bool) $config['features']['footnote'] ?? TRUE;
    
    $this->images_folder = $config['images_folder'] ?? 'images';
    
    $this->html = '';
    $this->footnotes = '';
    
    $odt_file_absolute_path = \Drupal::service('file_system')->realpath($odt_file);

    $this->pre_parse_odt($odt_file_absolute_path);
  }

  /**
   * {@inheritdoc}
   */
  public function html() {
    return $this->html . $this->footnotes;
  }  

  /**
   * {@inheritdoc}
   */
  protected function is_image($file) {

    $image_extensions = ['jpg', 'jpeg', 'png', 'gif', 'svg', 'webp'];
    $ext = pathinfo($file, PATHINFO_EXTENSION);
    if (!in_array($ext, $image_extensions)) {
      return FALSE;
    }

    // ToDo: change mime_content_type to symphony MimeTypeGuesserInterface
    return (strpos(@mime_content_type($file), 'image') === 0);
  }

  /**
   * {@inheritdoc}
   */
  protected function copy_file($from, $to) {
    //ToDo: change to file.repository service
    $file = file_get_contents($from); 
  
    // Load the service statically. 
    $filename = \Drupal::service('file_system')->saveData($file, $to, FileSystemInterface::EXISTS_REPLACE);
  
    return ($filename) ? \Drupal::service('file_url_generator')->generateAbsoluteString($filename) : FALSE;
  }

  /**
   * Function that preparses the XML and push it to odt2html()
   * ToDo: split this function to prepare_odt and convert_odt
   *
   */
  protected function pre_parse_odt($odt_file, $xml_string = NULL) {
    if ($xml_string === NULL) {
    	$xml_string = @file_get_contents('zip://' . $odt_file . '#content.xml');
      if (!$xml_string) {
        $this->messenger()->addError('Unable to read file contents.');

        return FALSE;
      }
    }
    
    // Remove page numbers from TOC
    $xml_string = preg_replace('/[ \t]*<text:tab\/>\d{1,}/isu', '', $xml_string);
    
    // Replace invalid links to h1,h2,h3 & etc. anchors
    //href="#\d{1,}\.([\w\t\d ]){2,}\|outline
    
    preg_match_all("/<text:bookmark-start text:name=\"([A-Za-z0-9_-]+)\"\/>(.*?)<text:bookmark-end text:name=\"([A-Za-z0-9_-]+)\"\/>/isu", $xml_string, $raw_bookmarks);
    
    $bookmarks = [];
    if (!empty($raw_bookmarks) && is_array($raw_bookmarks)) {
      foreach ($raw_bookmarks[2] as $key => $header) {
      	$bookmarks[$header.'|outline'] = '#'.$raw_bookmarks[1][$key];
      }
    }
    
    $xml_string = preg_replace_callback("/#\d{1,}\.(([\w\t\d ]){2,}\|outline)/imu", function ($match) use ($bookmarks) {
      return $bookmarks[ $match[1] ];
    }, $xml_string);
  
    // Revise this:
    // Convert breaks to br's
    //$xml_string = str_replace('<text:line-break/>', '<br />', $xml_string);
    
    return $this->odt2html($odt_file, $xml_string);
  }

  /**
   * Function that parses the XML and outputs HTML. If $xml is not provided,
   * extract content.xml from $odt_file
   */
  protected function odt2html($odt_file, $xml_string = NULL) {
    // ToDo: move to protected variable?
    $xml = new XMLReader();
  
    if ($xml_string === NULL) {
      if (@$xml->open('zip://' . $odt_file . '#content.xml') === FALSE) {
        $this->messenger()->addError('Unable to read file contents.');

        return FALSE;
      }
    } else {
      if (@$xml->xml($xml_string) === FALSE) {
        $this->messenger()->addError('Invalid file contents.');

        return FALSE;
      }
    }
  
    // Now, convert the xml from a string to an 
    $elements_tree = [];
  
    static $styles = ['Quotations' => ['tags' => ['blockquote']]];
  
    $translation_table = [];
  
    // removed by dillix
    // ToDo: do we need to reurn this here? See below: case 'draw:frame'
    //$translation_table['draw:frame'] = 'div class="odt-frame"';
  
    if ($this->features['list']) {
      $translation_table['text:list'] = 'ul';
      $translation_table['text:list-item'] = 'li';
    }
    else {
      $translation_table['text:list'] = FALSE;
    }

    if ($this->features['table']) {
      $translation_table['table:table'] = 'table class="table table-condensed table-bordered"';
      $translation_table['table:table-row'] = 'tr';
      $translation_table['table:table-cell'] = 'td';
    }
    else {
      $translation_table['table:table'] = FALSE;
    }

    if ($this->features['toc']) {
      $translation_table['text:table-of-content'] = 'div class="odt-table-of-contents"';
    }
    else {
      $translation_table['text:table-of-content'] = FALSE;
    }

    while ($xml->read()) {
      $opened_tags = []; //This array will contain the HTML tags opened in every iteration
  
      if ($xml->nodeType === XMLReader::END_ELEMENT) {
        // Handle a closing tag
        if (empty($elements_tree)) {
          continue;
        }

        do {
          $element = array_pop($elements_tree);
          if ($element && $element['tags']) {
            //Close opened tags
            $element['tags'] = array_reverse($element['tags']);
            foreach ($element['tags'] as $HTML_tag) {
                //$this->html .= "<font style='color:red' title='Closing $HTML_tag, from $element[name]. Current element is " .($xml->name). "'>©</font>";
                $HTML_tag = current(explode(' ', $HTML_tag));
                $this->html .= '</' . $HTML_tag . '>';
                //print($HTML_tag);
            }
          }
        } while ($xml->name !== $element['name'] && $element); //Close every opened tags. This should also handle malformed XML files
        continue;
      }
      elseif (in_array($xml->nodeType, [XMLReader::ELEMENT, XMLReader::TEXT, XMLReader::SIGNIFICANT_WHITESPACE])) {
        // Handle tags
        switch ($xml->name) {
          // Text
          case '#text':
            //$this->html .= str_replace(array("<",">"),array("&lt;","&gt;"), $xml->value);
            $this->html .= htmlspecialchars($xml->value);
            break;
  
          // Title
          // ToDo: Recheck this parser
          case 'text:h':
            if ($this->features['header']) {
              $n = $xml->getAttribute('text:outline-level');
              if ($n > 6) $n = 6;
              $opened_tags[] = 'h' . $n;
              $bookmark = '';
              
              // Bookmarks when table of contents exists in odt
              if ($this->features['bookmark']) {
                $child_xml_string = $xml->readInnerXML();
                if (!empty($child_xml_string)) {
                	if (preg_match('/text:name="([A-Za-z0-9_-]+)"/', $child_xml_string, $bookmark_id)) {
                	  $bookmark=' id="' . $bookmark_id[1] . '"';
                	}
                }
              }
  
              $this->html .= "\n\n<h$n$bookmark>";
            }
            else {
              $xml->next();
              break;
            }
            break;
  
          // Paragraph
          case 'text:p':
            //Just convert odf <text:p> to html <p>
            $tags = @$styles[$xml->getAttribute('text:style-name')]['tags'];
            if (!($tags && !in_array('blockquote', $tags))) {
              // Do not print a <p> immediatly after or before a <blockquote>
              $opened_tags[] = 'p';
              $this->html .= "\n<p>";
            }
            break;

          // ToDo: Recheck this parser  
          case 'text:a':
            if ($this->features['link']) {
              $href = $xml->getAttribute('xlink:href');
              $opened_tags[] = 'a';
              $this->html .= '<a href="' . $href . '">';
            }
            else {
              $xml->next();
              break;
            }
            break;
  
          case 'draw:frame':
            if ($this->features['frame']) {
              $frame_type = $xml->getAttribute('text:anchor-type');
              if ($frame_type == 'as-char') {
                $opened_tags[] = 'span';
                $this->html .= '<span class="odt-frame">';
              }
              else {
                $opened_tags[] = 'div';
                $this->html .= '<div class="odt-frame">';
              }              
            }
            else {
              $xml->next();
              break;
            }
            break;
  
          case 'draw:image':
            if ($this->features['image']) {
              $image_file = 'zip://' . $odt_file . '#' . $xml->getAttribute("xlink:href");
              if (isset($this->images_folder) && is_dir($this->images_folder)) {
                if ($this->is_image($image_file)) {
                  $image_to_save = $this->images_folder . '/' . basename($image_file);
                  if (!($src = $this->copy_file($image_file, $image_to_save))) {
                    $this->messenger()->addError('Unable to move image file.');
                    break;
                  }
                  else {
                    // ToDo: make config option for size of inline images
                    $image_size = getimagesize($image_to_save);
                    if ((int) $image_size[0] < 128) {
                      $class = 'img-inline';
                    }
                    else {
                      $class = 'img-responsive';
                    }
                  }
                }
                else {
                  $this->messenger()->addError('Found invalid image file.');
                  break;
                }
              }
              else {
                $this->messenger()->addError('Unable to save the image. Creating a data URL. Image saved directly in the body.');
                $src = 'data:image;base64,' . base64_encode(file_get_contents($image_file));
              }
              $this->html .= '<img src="' . $src . '" class="' . $class .'" />';
            }
            else {
              $xml->next();
              break;              
            }
            break;

          case "text:line-break"://br
            $this->html .= "\n";//<br />
            break;
  
          case "style:style":
            $name = $xml->getAttribute('style:name');
            $parent = $xml->getAttribute('style:parent-style-name');
            if (array_key_exists($parent, $styles)) {
              $styles[$name] = $styles[$parent]; //Not optimal
            }
  
            if ($xml->isEmptyElement) {
              break; //We can't handle that at the moment
            }


            while ($xml->read() && ($xml->name != 'style:style' || $xml->nodeType != XMLReader::END_ELEMENT)) {
              if ($xml->name == 'style:text-properties') {
                //Creates the style and add <i> to its tags
                if ($xml->getAttribute('fo:font-style') == 'italic') {
                  $styles[$name]['tags'][] = 'i';
                }

                //Creates the style and add <b> to its tags
                if ($xml->getAttribute('fo:font-weight') == 'bold') {
                  $styles[$name]['tags'][] = 'b';
                }

                //Creates the style and add <b> to its tags
                if ($xml->getAttribute('style:text-underline-style') == 'solid') {
                  $styles[$name]['tags'][] = 'u';
                }
              }
            }
            break;

          // ToDo: Recheck this parser 
          case 'text:note':
            if ($this->features['note']) {
              $note_id = $xml->getAttribute('text:id');
              $note_name = 'Note';

              // Read one tag and stop on </style:style>              
              while ($xml->read() &&  ($xml->name != 'text:note' || $xml->nodeType != XMLReader::END_ELEMENT)) {
                if ($xml->name == 'text:note-citation' && $xml->nodeType == XMLReader::ELEMENT) {
                  $note_name = $xml->readString();
                }
                elseif ($xml->name == 'text:note-body' && $xml->nodeType == XMLReader::ELEMENT) {
                  $note_content = $this->odt2html($odt_file, $xml->readOuterXML());
                }
              }
    
              $this->html .= '<sup><a href="#odt-footnote-' . $note_id . '" class="odt-footnote-anchor" name="anchor-odt-' . $note_id . '">$note_name</a></sup>';
    
              $this->footnotes .= "\n" . '<div class="odt-footnote" id="odt-footnote-' . $note_id . '">';
              $this->footnotes .= '<a class="footnote-name" href="#anchor-odt-' . $note_id . '">' . $note_name . ' .</a> ';
              $this->footnotes .= $note_content;
              $this->footnotes .= '</div>' . "\n";
            } else {
              $xml->next();
              break;              
            }
            break;
  
          // ToDo: Recheck this parser 
          case 'office:annotation':
            if ($this->features['annotation']) {
              $annotation_id = (isset($annotation_id)) ? $annotation_id + 1 : 1;
              $annotation_content = '';
              $annotation_creator = $this->t('Anonymous');
              $annotation_date = '';
              do {
                $xml->read();
                if ($xml->name == 'dc:creator' && $xml->nodeType == XMLReader::ELEMENT) {
                  $annotation_creator = $xml->readString();                  
                }
                elseif($xml->name == 'dc:date' && $xml->nodeType == XMLReader::ELEMENT) {
                  // ToDo: Add date format to module configuration
                  $annotation_date = date("jS \of F Y, H\h i\m", strtotime($xml->readString()));
                }
                elseif ($xml->nodeType == XMLReader::ELEMENT) {
                  $annotation_content .= $xml->readString();
                  $xml->next();
                }
              } while (!($xml->name === 'office:annotation' && $xml->nodeType === XMLReader::END_ELEMENT));
              //End of the note
    
              $this->html .= '<sup><a href="#odt-annotation-' . $annotation_id . '" name="anchor-odt-annotation-' . $annotation_id . '" title="Annotation (' . $annotation_creator . ')">(' . $annotation_id . ')</a></sup>';
              $this->footnotes .= "\n" . '<div class="odt-annotation" id="odt-annotation-' . $annotation_id . '" >';
              $this->footnotes .= '<a class="annotation-name" href="#anchor-odt-annotation-' . $annotation_id . '"> (' . $annotation_id . ')&nbsp;</a>';
              $this->footnotes .= "\n" . '<b>' . $annotation_creator . ' (<i>' . $annotation_date . '</i>)</b> :';
              $this->footnotes .= "\n" . '<div class="odt-annotation-content">' . $annotation_content . '</div>';
              $this->footnotes .= '</div>' . "\n";
            } 
            else {
              $xml->next();
              break;              
            }
            break;
  
          default:
            if (array_key_exists($xml->name, $translation_table)) {
              if ($translation_table[$xml->name] === FALSE) {
                $xml->next();
                break;
              }
              $tag = explode(' ', $translation_table[$xml->name], 1);
              //$tag[0] is the tag name, other indexes are attributes
              $opened_tags[] = $tag[0];
              $this->html .= "\n<" . $translation_table[$xml->name] . '>';
            }
        }
      }
  
      if ($xml->nodeType === XMLReader::ELEMENT && !($xml->isEmptyElement)) { //Opening tag
        $current_element_style = $xml->getAttribute('text:style-name');
        if ($current_element_style && isset($styles[$current_element_style])) {
          //Styling tags management
          foreach ($styles[$current_element_style]['tags'] as $HTML_tag) {
            $this->html .= '<' . $HTML_tag . '>';
            $opened_tags[] = $HTML_tag;
          }
        }
        $elements_tree[] = [
          'name' => $xml->name,
          'tags' => $opened_tags
        ];
      }
    }
  }
}
